import yaml
import os
import click
import zipfile
import urllib
import logging
import jsonpointer
import jq
import glob2 as glob

log = logging.getLogger(__name__)


def leaf_iterator(jsonable):
    allleafs = jq.jq('leaf_paths').transform(jsonable, multiple_output = True)
    leafpointers = [jsonpointer.JsonPointer.from_parts(x) for x in allleafs]
    for x in leafpointers:
        yield x,x.get(jsonable)

def discover_initfiles(initdata,sourcedir):
    '''inspect sourcedir, first tries exact path match, and then (possbly recursive) glob'''
    log.info('inspecting %s to discover referenced input files',sourcedir)

    # filled_initdata = copy.deepcopy(initdata)
    for pointer,value in leaf_iterator(initdata):
        if type(value) not in [str,unicode]: continue
        within_sourcedir = os.path.join(sourcedir,value)
        globresult = glob.glob(os.path.join(sourcedir,value))
        if os.path.exists(within_sourcedir):
            pointer.set(initdata,within_sourcedir)
        elif globresult:
            pointer.set(initdata,globresult)
    return initdata

def getinit_data(initfiles, parameters):
    '''
    get initial data from both a list of files and a list of 'pname=pvalue'
    strings as they are passed in the command line <pvalue> is assumed to be a
    YAML parsable string.
    '''
    initdata = {}
    for initfile in initfiles:
        log.info('loading initialization data from file %s',initfile)
        initdata.update(**yaml.load(open(initfile)))

    for x in parameters:
        key, value = x.split('=')
        initdata[key] = yaml.load(value)
    return initdata


def prepare_workdir_from_archive(workdir, inputarchive):
    if os.path.exists(workdir):
        raise click.exceptions.ClickException(click.style(
            "workdirectory exists and input archive give. Can't have both", fg='red'))
    initdir = os.path.join(workdir,'init')
    os.makedirs(initdir)
    localzipfile = '{}/inputarchive.zip'.format(workdir)
    urllib.urlretrieve(inputarchive, localzipfile)
    with zipfile.ZipFile(localzipfile) as zf:
        zf.extractall(path=initdir)
    os.remove(localzipfile)
    return initdir


def setupbackend_fromstring(backend, name = 'backendname', cacheconfig=None):
    import backends.packtivitybackend as pb
    backend = pb.PacktivityBackend(packtivity_backendstring = backend, cacheconfig = cacheconfig)
    return backend



import federatedbackend
import json
import time
import os
import logging
import jq
import jsonpointer
import checksumdir

from yadage.helpers import get_obj_id
from trivialbackend import TrivialProxy, TrivialBackend

log = logging.getLogger(__name__)

class CachedBackend(federatedbackend.FederatedBackend):
    '''
    A caching backend that takes an existing backend and a cache configuration.
    If at submission time a cache for a given task is valid, the result will be
    returned directly via a trivial proxy. If not submission proceeds as normal
    '''

    def __init__(self, backend, cacheconfig):
        super(CachedBackend, self).__init__({
            'cache': TrivialBackend(),
            'primary': backend
        })
        configparts  = cacheconfig.split(':')
        if len(configparts) == 1:
            log.info('default caching strategy')
            self.cache = CacheBuilder(configparts[0])
        else:
            strategy, configfile = configparts
            if strategy == 'resultexists':
                log.info('resultexists caching strategy')
                self.cache = ResultFilesExistCache(configfile)
            else:
                raise RuntimeError('unknown caching config')

    def ready(self, proxy):
        isready = super(CachedBackend, self).ready(proxy)
        #if this a genuinely new result, cache it under the task cache id
        if isready and type(proxy) is not TrivialProxy:
            result = super(CachedBackend, self).result(proxy)
            status = super(CachedBackend, self).successful(proxy)
            if not self.cache.cacheexists(proxy.cacheid):
                self.cache.cacheresult(proxy.cacheid, status, result)
        return isready

    def routedsubmit(self, task):
        cached = self.cache.cacheddata(task)
        if cached:
            log.info('use cached result for task: %s',task.name)
            return TrivialProxy(status=cached['status'], result = cached['result'])
        else:
            log.info('do proper submit for task: %s',task.name)
            #create id for this task using the cache builder, with which
            #we will store the result with once it's ready
            cacheid = self.cache.cacheid(task)
            primaryproxy = self.backends['primary'].submit(task.spec, task.attributes, task.context)
            primaryproxy.cacheid = cacheid
            return primaryproxy

    def routeproxy(self, proxy):
        if type(proxy) == TrivialProxy:
            return 'cache'
        else:
            return 'primary'


class CacheBuilder(object):
    def __init__(self, cachefile):
        self.cachefile = cachefile
        if not os.path.exists(self.cachefile):
            self.cache = {}
        else:
            log.info('reading cache from %s',self.cachefile)
            self.cache = json.load(open(self.cachefile))

    def __del__(self):
        log.info('writing cache to %s',self.cachefile)
        json.dump(self.cache, open(self.cachefile, 'w'), indent=4, sort_keys=True)

    def remove(self,cacheid):
        self.cache.pop(cacheid)

    def cacheresult(self, cacheid, status, result):
        log.info('caching result for cacheid: %s',cacheid)
        log.info('caching result for process: %s',self.cache[cacheid]['task']['spec']['process'])
        self.cache[cacheid]['result'] = {
            'status': 'SUCCESS' if status else 'FAILED',
            'result': result,
            'cachingtime': time.time(),
            'checksums': None
        }
        if 'context' in self.cache[cacheid]['task']:
            checksums = [checksumdir.dirhash(d) for d in self.cache[cacheid]['task']['context']['depwrites']]
            log.info('compute checksums for %s',self.cache[cacheid]['task']['context']['depwrites'])
            log.info('checksums are %s',checksums)
            self.cache[cacheid]['result']['checksums'] = checksums

    def cachedresult(self,cacheid, silent = True):
        '''
        returns the cached result. when silent = True the mthod exits gracefully
        and returns None
        '''
        if silent:
            if not self.cacheexists(cacheid): return None
        return self.cache[cacheid]['result']

    def cacheid(self,task):
        return get_obj_id(task, method = 'jsonhash')

    def cacheexists(self,cacheid):
        return cacheid in self.cache and 'result' in self.cache[cacheid]

    def cachevalid(self, cacheid):
        return True

    def cacheddata(self, task, remove_invalid = True):
        '''
        returns results from a valid cache entry if it exists, else None
        if remove_invalid = True, also removes invalid cache entries from cache
        '''
        cacheid = self.cacheid(task)
        #register this task with the cacheid if we don't know about it yet
        log.info('checking cache for task %s',task.name)
        if cacheid not in self.cache:
            self.cache[cacheid] = {'task' : task.json()}
        if not self.cacheexists(cacheid):
            log.info('cache non-existent for task %s (%s)',task.name,cacheid)
            return None
        if not self.cachevalid(cacheid):
            log.info('cache non-valid for task %s (%s)',task.name,cacheid)
            self.remove(cacheid)
            self.cache[cacheid] = {'task' : task.json()}
            return None
        #return a cached result if we have one if not, return None
        result =  self.cachedresult(cacheid, silent = True)
        log.info('returning cached result %s',result)
        return result

class ResultFilesExistCache(CacheBuilder):
    '''
    This is a cache that attempts checks for file paths in result object
    leaves and checks if they already exist. If yes
    '''
    def __init__(self, cachefile):
        super(ResultFilesExistCache, self).__init__(cachefile)

    def remove(self,cacheid):
        import shutil
        task = self.cache[cacheid]['task']
        log.info('removing cache entry %s',cacheid)
        workdir = task['context']['readwrite'][0]
        log.info('deleting rw location %s',workdir)
        if os.path.exists(workdir):
            shutil.rmtree(workdir)
        super(ResultFilesExistCache, self).remove(cacheid)

    def cachevalid(self, cacheid):
        task = self.cache[cacheid]['task']
        result = self.cache[cacheid]['result']['result']
        if task['type'] == 'initstep':
            return True


        #check if dependent state is still the same

        stored_checksums = self.cache[cacheid]['result']['checksums']
        checksums_now = [checksumdir.dirhash(d) for d in task['context']['depwrites']]
        log.info('checksums comparison: %s',checksums_now == stored_checksums)
        if not checksums_now == stored_checksums:
            log.info('cache invalid due to changed input state')
            log.info(task['context']['depwrites'])
            log.info(stored_checksums)
            log.info(checksums_now)
            return False

        #check if our result fragments are still there
        rwloc = task['context']['readwrite'][0]
        resultleafs = jq.jq('leaf_paths').transform(result, multiple_output=True)
        resultleafs = [jsonpointer.JsonPointer.from_parts(x).resolve(result) for x in resultleafs]
        for x in resultleafs:
            if rwloc in x:
                exists = os.path.exists(x)
                if not exists:
                    log.info('cache invalid due to missing data fragments')
                    return False
        log.info('all file refs in result exist already.')
        return True



import jsonpointer
import networkx as nx
import packtivity.statecontexts.poxisfs_context as statecontext


def select_rule(workflow, offset, name):
    for x in workflow.applied_rules:
        if x.offset == offset and x.rule.name == name:
            return x
    raise RuntimeError('rule not found')


def stepsofrule(workflow, offset, name):
    rule = select_rule(workflow, offset, name)
    path = '/'.join([rule.offset, rule.rule.name])
    p = jsonpointer.JsonPointer(path)
    return rule, [x['_nodeid'] for x in p.resolve(workflow.stepsbystage)]


def reset_node_state(node):
    node.submit_time = None
    node.ready_by_time = None
    node.resultproxy = None
    node.backend = None


def reset_step(workflow, step):
    s = workflow.dag.getNode(step)
    reset_node_state(s)
    try:
        statecontext.reset_state(s.task.context)
    except AttributeError:
        pass


def reset_steps(workflow, steps):
    for s in steps:
        reset_step(workflow, s)


def collective_downstream(workflow, steps):
    downstream = set()
    for step in steps:
        for x in nx.descendants(workflow.dag, step):
            downstream.add(x)
    return list(downstream)


def reset_state(workflow, offset, name):
    rule, steps = stepsofrule(workflow, offset, name)
    toreset = steps + collective_downstream(workflow, steps)
    reset_steps(workflow, toreset)



#!/usr/bin/env python
import logging
import click
import os
import manualutils
import yadage.workflow_loader
import packtivity.statecontexts.poxisfs_context as statecontext
import clihelpers
import serialize
import adage
import reset as yr

log = logging.getLogger(__name__)


@click.group()
def mancli():
    pass


@mancli.command()
@click.option('-s', '--statefile', default='yadage_wflow_state.json')
@click.option('-b', '--backendfile', default='yadage_backend_state.json')
@click.option('-t', '--toplevel', default=os.getcwd())
@click.option('-a', '--inputarchive', default=None)
@click.option('--parameter', '-p', multiple=True)
@click.argument('workdir')
@click.argument('workflow')
@click.argument('initfiles', default='')
def init(workdir, workflow, initfiles, statefile, backendfile, toplevel, parameter, inputarchive):
    workflow_def = yadage.workflow_loader.workflow(
        toplevel=toplevel,
        source=workflow
    )

    rootcontext = statecontext.make_new_context(workdir)
    workflow = yadage.yadagemodels.YadageWorkflow.createFromJSON(
        workflow_def, rootcontext)

    initdata = clihelpers.getinit_data(initfiles, parameter)
    workflow.view().init(initdata)

    click.secho('initialized workflow', fg='green')

    yadagedir = manualutils.get_yadagedir(workdir)
    os.makedirs(yadagedir)

    statefile = '{}/{}'.format(manualutils.get_yadagedir(workdir), statefile)
    backendfile = '{}/{}'.format(manualutils.get_yadagedir(workdir),
                                 backendfile)

    click.secho('statefile at {}'.format(statefile))
    serialize.snapshot(
        workflow,
        statefile,
        backendfile
    )


@mancli.command()
@click.argument('workdir')
@click.argument('name', default='')
@click.option('-o', '--offset', default='')
@click.option('-s', '--statefile', default='yadage_wflow_state.json')
@click.option('-b', '--backendfile', default='yadage_backend_state.json')
@click.option('-v', '--verbosity', default='ERROR')
def apply(workdir, name, offset, statefile, backendfile, verbosity):
    logging.basicConfig(level=getattr(logging, verbosity))
    with manualutils.workflowctx(workdir, statefile, backendfile) as (backend, workflow):
        if not name:
            click.secho('Applicable Rules: ', fg='blue')
            for x in manualutils.applicable_rules(workflow):
                click.secho('{}/{}'.format(x.offset, x.rule.name))
            return
        rule = workflow.view(offset).getRule(name)
        if not rule:
            click.secho(
                'No such rule, pick one of the applicable below:', fg='red')
            for x in manualutils.applicable_rules(workflow):
                click.secho('{}/{}'.format(x.offset, x.rule.name))
            return
        if not rule.applicable(workflow):
            click.secho('Rule is not applicable.', fg='red')
            return
        workflow.rules.remove(rule)
        rule.apply(workflow)
        workflow.applied_rules.append(rule)


@mancli.command()
@click.argument('workdir')
@click.argument('name')
@click.option('-o', '--offset', default='')
@click.option('-s', '--statefile', default='yadage_wflow_state.json')
@click.option('-b', '--backendfile', default='yadage_backend_state.json')
@click.option('-v', '--verbosity', default='ERROR')
def preview(workdir, name, offset, statefile, backendfile, verbosity):
    with manualutils.workflowctx(workdir, statefile, backendfile) as (backend, workflow):
        new_rules, new_nodes = manualutils.preview_rule(workflow, name, offset)
        click.secho('Preview of Stage: # new rules: {} # new nodes {}'.format(
            len(new_rules), len(new_nodes)))
        for n in new_nodes:
            click.secho(
                '-> new node "{}" with {} upstream dependencies'.format(n['name'], len(n['parents'])))


@mancli.command()
@click.argument('workdir')
@click.option('-s', '--statefile', default='yadage_wflow_state.json')
@click.option('-b', '--backendfile', default='yadage_backend_state.json')
@click.option('-v', '--verbosity', default='ERROR')
def step(workdir, statefile, backendfile, verbosity):
    logging.basicConfig(level=getattr(logging, verbosity))
    with manualutils.workflowctx(workdir, statefile, backendfile) as (backend, workflow):
        extend_decider, submit_decider = yadage.interactive.interactive_deciders()
        coroutine = adage.adage_coroutine(
            backend, extend_decider, submit_decider)
        coroutine.next()  # prime the coroutine....
        coroutine.send(workflow)
        try:
            click.secho('try stepping workflow')
            coroutine.next()
        except StopIteration:
            manualutils.finalize_manual(workdir, workflow)


@mancli.command()
@click.argument('workdir')
@click.argument('name')
@click.option('-s', '--statefile', default='yadage_wflow_state.json')
@click.option('-b', '--backendfile', default='yadage_backend_state.json')
@click.option('-o', '--offset', default='')
def reset(workdir, statefile, backendfile, offset, name):
    with manualutils.workflowctx(workdir, statefile, backendfile) as (backend, workflow):
        yr.reset_state(workflow, offset, name)

if __name__ == '__main__':
    mancli()



import yaml
import json
import click
import tempfile
import shutil
import yadagemodels
import visualize
import logging
from yadagestep import outputReference
from handlers.expression_handlers import handlers as exh
from helpers import set_trivial_backend


def process(x, dag):
    if type(x) == dict:
        for k, v in x.iteritems():
            x[k] = process(v, dag)
        return x
    elif type(x) == list:
        for i, e in enumerate(x):
            x[i] = process(e, dag)
        return x
    elif type(x) == outputReference:
        return x.pointer.resolve(dag.getNode(x.stepid).result)
    else:
        return x


def printRef(ref, dag, indent=''):
    click.secho('{}name: {} position: {}, value: {}, id: {}'.format(
        indent,
        dag.getNode(ref.stepid).name,
        ref.pointer.path,
        ref.pointer.resolve(dag.getNode(ref.stepid).result),
        ref.stepid
    ),
        fg='cyan')


def wflow_with_trivial_backend(instance,results):
    wflow = yadagemodels.YadageWorkflow.fromJSON(
        json.load(open(instance))
    )
    set_trivial_backend(wflow.dag, json.load(open(results)))
    return wflow

@click.group()
def utilcli():
    pass


@utilcli.command()
@click.argument('instance')
@click.argument('results')
@click.argument('selection')
@click.option('-v', '--verbosity', default='INFO')
def testsel(instance, results, selection,verbosity):


    logging.basicConfig(level=getattr(logging, verbosity), format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


    wflow = wflow_with_trivial_backend(instance,results)

    selresult = exh['stage-output-selector'](wflow.view(), yaml.load(selection))

    if not selresult:
        click.secho('Bad selection', fg='red')
        return

    click.secho(json.dumps(
        process(selresult, wflow.dag),
        sort_keys=True,
        indent=4,
        separators=(',', ': ')),
        fg='green'
    )


@utilcli.command()
@click.argument('instance')
@click.argument('results')
@click.argument('vizpdf')
def viz(instance, results, vizpdf):
    wflow = wflow_with_trivial_backend(instance,results)

    dirpath = tempfile.mkdtemp()
    visualize.write_prov_graph(dirpath, wflow)
    shutil.copy('{}/yadage_workflow_instance.pdf'.format(dirpath), vizpdf)
    shutil.rmtree(dirpath)


if __name__ == '__main__':
    utilcli()



import logging
import utils
import itertools

from expression_handlers import handlers as exprhandlers
from yadage.yadagestep import yadagestep, initstep, outputReference
from yadage.yadagemodels import jsonStage
import packtivity.statecontexts.poxisfs_context as statecontext

log = logging.getLogger(__name__)

handlers, scheduler = utils.handler_decorator()

# A scheduler does the following things:
#   - attached new nodes to the DAG
# - for each added step
#     - the step is given a name
#     - the step attributes are determined using the scheduler spec and context
#     - a list of used inputs (in the form of [stepname,outputkey,index])


def select_parameter(stageview, parameter):
    if type(parameter) is not dict:
        value = parameter
    else:
        handler = exprhandlers[parameter['expression_type']]
        value = handler(stageview, parameter)
    return value


def finalize_value(stage, step, value, context):
    '''
    finalize a value by recursively resolving references and
    interpolating with the context when necessary
    '''
    if type(value) == outputReference:
        step.used_input(value)
        v = value.pointer.resolve(stage.view.dag.getNode(value.stepid).result)
        return finalize_value(stage, step, v, context)
    if type(value) == list:
        return [finalize_value(stage, step, x, context) for x in value]
    if type(value) in [str, unicode]:
        return value.format(**context)
    return value


def finalize_input(stage, step, json, context):
    '''
    evaluate final values of parameters by either resolving a
    reference to a upstream output or evaluating a static
    reference from the template (possibly string-interpolated)
    '''

    context = context.copy()
    context['workdir'] = context['readwrite'][0]
    result = {}
    for k, v in json.iteritems():
        if type(v) is not list:
            result[k] = finalize_value(stage, step, v, context)
        else:
            result[k] = [finalize_value(
                stage, step, element, context) for element in v]

    return result


def step_or_init(name, spec, context):
    if 'step' in spec:
        stepcontext = statecontext.make_new_context(name, context, subdir=True)
        return yadagestep(name=name, spec=spec['step'], context=stepcontext)
    elif 'workflow' in spec:
        return initstep('init {}'.format(name))


def addStepOrWorkflow(name, stage, step, spec):
    if type(step) == initstep:
        newcontext = statecontext.make_new_context(name, stage.context)
        subrules = [jsonStage(yml, newcontext)
                    for yml in spec['workflow']['stages']]
        stage.addWorkflow(subrules, initstep=step)
    else:
        stage.addStep(step)


def get_parameters(spec):
    return {x['key']: x['value']for x in spec['parameters']}


@scheduler('singlestep-stage')
def singlestep_stage(stage, spec):
    '''
    a simple state that adds a single step/workflow. The node is attached
    to the DAG based on used upstream outputs
    '''
    log.debug('scheduling singlestep stage with spec:\n%s', spec)

    step = step_or_init(name=stage.name, spec=spec, context=stage.context)
    ctx = step.context if hasattr(step, 'context') else stage.context

    parameters = {
        k: select_parameter(stage.view, v) for k, v in get_parameters(spec).iteritems()
    }
    finalized = finalize_input(stage, step, parameters, ctx)

    addStepOrWorkflow(stage.name, stage, step.s(**finalized), spec)


def scatter(parameters, scatter):
    '''
    this method turns a parameter set and scatter definition into a list
    of single parameter sets.
    '''
    commonpars = parameters.copy()
    to_scatter = {}
    for scatpar in scatter['parameters']:
        to_scatter[scatpar] = commonpars.pop(scatpar)

    singlesteppars = []
    if scatter['method'] == 'zip':
        keys, zippable = zip(
            *[(k, v) for i, (k, v) in enumerate(to_scatter.iteritems())])

        for zipped in zip(*zippable):
            individualpars = dict(zip(keys, zipped))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]

    if scatter['method'] == 'cartesian':
        for what in itertools.product(*[to_scatter[k] for k in scatter['parameters']]):
            individualpars = dict(zip(scatter['keys'], what))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]
    return singlesteppars


@scheduler('multistep-stage')
def multistep_stage(stage, spec):
    '''
    a stage that attaches an array of nodes to the DAG. The number of nodes
    is determined by a scattering recipe. Currently two algs are supported
    'zip': one or more arrays of length n are iterated through in lock-step.
           n nodes are added to the DAG where the  parameters values are set to
           the values in the iteration
    'cartesian': a cartesian product of a number of arrays (possibly different sizes)
                 adds n1 x n2 x ... nj nodes.
    Nodes are attached to the DAG based on used upstream inputs
    '''
    log.debug('scheduling multistep stage with spec:\n%s', spec)
    parameters = {
        k: select_parameter(stage.view, v) for k, v in get_parameters(spec).iteritems()
    }
    singlesteppars = scatter(parameters, spec['scatter'])
    for i, pars in enumerate(singlesteppars):
        singlename = '{}_{}'.format(stage.name, i)
        step = step_or_init(name=singlename, spec=spec, context=stage.context)
        ctx = step.context if hasattr(step, 'context') else stage.context
        ctx = ctx.copy()
        ctx.update(index=i)
        finalized = finalize_input(stage, step, pars, ctx)
        addStepOrWorkflow(singlename, stage, step.s(**finalized), spec)



import json
import click
import time
import copy
import yadage.workflow_loader
import yadage.yadagemodels
import yadage.visualize
import yadage.interactive
import serialize
import adage.visualize as av

from contextlib import contextmanager


@contextmanager
def workflowctx(workdir, statefile, backendfile):
    statefile = '{}/{}'.format(get_yadagedir(workdir), statefile)
    backendfile = '{}/{}'.format(get_yadagedir(workdir), backendfile)
    backend, workflow = load_state(statefile)

    yield backend, workflow

    serialize.snapshot(
        workflow,
        statefile,
        backendfile
    )
    av.save_dot(
        av.colorize_graph_at_time(workflow.dag, time.time()).to_string(),
        '{}/{}'.format(get_yadagedir(workdir),
                       'adage.png'),
        'png'
    )


def decide_rule(rule, state):
    click.secho('we could extend DAG with rule', fg='blue')
    click.secho('rule: {}/{} ({})'.format(rule.offset,
                                          rule.rule.name, rule.identifier))
    shall = raw_input(click.style(
        "Shall we? (y/N) ", fg='blue')).lower() == 'y'
    if shall:
        click.secho('ok we will extend.', fg='green')
    else:
        click.secho('maybe another time...', fg='yellow')
    return shall


def decide_step(dag, nodeobj):
    print 'we could submit a DAG node (id: {}) DAG is: {}'.format(nodeobj, dag)
    shall = raw_input(click.style("Shall we? (y/N) ",
                                  fg='magenta')).lower() == 'y'
    if shall:
        click.secho('ok we will submit.', fg='green')
    else:
        click.secho('will not submit for now...', fg='yellow')
    return shall


def custom_decider(decide_func):
    # we yield until we receive some data via send()
    def decider():
        data = yield
        while True:
            data = yield decide_func(*data)
    return decider


def get_yadagedir(workdir):
    return '{}/_yadage'.format(workdir)


class VariableProxy():
    @staticmethod
    def fromJSON(data):
        import packtivity.asyncbackends
        import yadage.backends.packtivitybackend
        if data['proxyname']=='InitProxy':
            return yadage.backends.packtivitybackend.InitProxy.fromJSON(data)
        elif data['proxyname']=='CeleryProxy':
            return packtivity.asyncbackends.CeleryProxy.fromJSON(data)
        else:
            raise RuntimeError('only celery support for now...')

def load_state(statefile):
    from clihelpers import setupbackend_fromstring

    backend = setupbackend_fromstring('celery')

    click.secho('loading state from {}'.format(statefile))
    workflow = yadage.yadagemodels.YadageWorkflow.fromJSON(
        json.load(open(statefile)),
        VariableProxy,
        backend
    )
    return backend, workflow

def finalize_manual(workdir, workflow):
    click.secho('workflow done.', fg='green')
    yadage.visualize.write_prov_graph(get_yadagedir(workdir), workflow)


def applicable_rules(workflow, justids=False):
    return [(x.identifier if justids else x) for x in workflow.rules if x.applicable(workflow)]


def preview_rule(wflow, name=None, offset='', identifier=None):
    newflow = yadage.yadagemodels.YadageWorkflow.fromJSON(
        copy.deepcopy(wflow.json())
    )

    if identifier:
        rule = newflow.view().getRule(identifier=identifier)
    else:
        rule = newflow.view().getRule(name, offset)

    if not rule.applicable(newflow):
        return

    rule.apply(newflow)
    newflow.rules.remove(rule)
    newflow.applied_rules.append(rule)

    existing_rules = [x.identifier for x in (
        wflow.rules + wflow.applied_rules)]
    existing_nodes = wflow.dag.nodes()

    new_rules = [{'name': x.rule.name, 'offset': x.offset}
                 for x in newflow.rules if x.identifier not in existing_rules]
    new_nodes = [{'name': newflow.dag.getNode(n).name, 'parents': newflow.dag.predecessors(
        n)} for n in newflow.dag.nodes() if n not in existing_nodes]
    return new_rules, new_nodes



import click


def decide_rule(rule, state):
    click.secho('we could extend DAG with rule', fg='blue')
    click.secho('rule: {}/{} ({})'.format(rule.offset,
                                          rule.rule.name, rule.identifier))
    resp = raw_input(click.style("Shall we? (y/N) ", fg='blue'))
    shall = resp.lower() == 'y'
    if shall:
        click.secho('ok we will extend.', fg='green')
    else:
        click.secho('maybe another time... your response: {}'.format(
            resp), fg='yellow')
    return shall


def decide_step(dag, nodeobj):
    click.echo(
        'we could submit a DAG node (id: {}) DAG is: {}'.format(nodeobj, dag))
    resp = raw_input(click.style("Shall we? (y/N) ", fg='magenta'))
    shall = resp.lower() == 'y'
    if shall:
        click.secho('ok we will submit.', fg='green')
    else:
        click.secho('will not submit for now... your response: {}'.format(
            resp), fg='yellow')
    return shall


def custom_decider(decide_func):
    # we yield until we receive some data via send()
    def decider():
        data = yield
        while True:
            data = yield decide_func(*data)
    return decider


def interactive_deciders():
    '''
    returns a tuple (extend,submit) of already-primed deciders for both
    extension and submission
    '''
    extend_decider = custom_decider(decide_rule)()
    extend_decider.next()  # prime decider

    submit_decider = custom_decider(decide_step)()
    submit_decider.next()  # prime decider

    return extend_decider, submit_decider
